from ...ir import Variable
from typing import List, TYPE_CHECKING
from PySide6.QtCore import Signal, Qt, QRectF, QPointF, Slot
from PySide6.QtGui import QPainter, QColor, QPen, QFont, QFontMetrics, QPainterPath, QPainterPathStroker
from PySide6.QtWidgets import QGraphicsItem, QGraphicsObject, QGraphicsSceneHoverEvent, QGraphicsSceneMouseEvent, QStyleOptionGraphicsItem, QWidget, QGraphicsPathItem
from typing import Union

if TYPE_CHECKING:
    from .node import GraphNode


class GraphEdge(QGraphicsObject):
    def __init__(self, ir: Variable, parent=None):
        super().__init__()
        self._ir: Variable = ir

        self.setAcceptHoverEvents(True)
        self.setZValue(0)

        self._hovered = False

        self._src_pts: List[QPointF] = []
        self._dst_pts: List[QPointF] = []
        self._path: QPainterPath = QPainterPath()

    @property
    def ir(self):
        return self._ir

    @Slot(list)
    def needUpdate(self, nodes: List['GraphNode']):
        assert len(nodes) == 1
        node = nodes[0]
        src_obj = [n.read_ext('bind_gnode') for n in self._ir.src if n.read_ext(
            'bind_gnode') is not None]
        dst_obj = [n.read_ext('bind_gnode') for n in self._ir.dst if n.read_ext(
            'bind_gnode') is not None]
        if self._ir.isInput:
            assert len(src_obj) == 0
            src_obj = [self._ir.read_ext('bind_gnode')]
            assert src_obj[0] is not None
        if self._ir.isOutput:
            dst_obj += [self._ir.read_ext('bind_gnode')]
            assert dst_obj[-1] is not None
        self.drawEdge(src_obj, dst_obj)
        if node not in src_obj + dst_obj:
            node.pos_move.disconnect(self.needUpdate)
            node.io_change.disconnect(self.needUpdate)

    def drawEdge(self, src_obj: List[QGraphicsItem], dst_obj: List[QGraphicsItem]):
        self._src_pts.clear()
        for o in src_obj:
            pos = o.pos()
            rect = o.boundingRect()
            self._src_pts.append(
                QPointF(rect.left() + rect.width() / 2, rect.bottom()) + pos)
            # print('src', o._ir.name, pos, rect, self._src_pts[-1])
        self._dst_pts.clear()
        for o in dst_obj:
            pos = o.pos()
            rect = o.boundingRect()
            self._dst_pts.append(
                QPointF(rect.left() + rect.width() / 2, rect.top()) + pos)
            # print('dst', o._ir.name, pos, rect, self._src_pts[-1])
        self._path.clear()
        for s in self._src_pts:
            for d in self._dst_pts:
                self._path.moveTo(s)
                c1 = QPointF(s.x(), (s.y() + d.y()) / 2)
                c2 = QPointF(d.x(), (s.y() + d.y()) / 2)
                self._path.cubicTo(c1, c2, d)
        self.prepareGeometryChange()
        self.update()

    def shape(self) -> QPainterPath:
        ps = QPainterPathStroker()
        pen = QPen()
        if self._hovered or self.isSelected():
            pen.setWidthF(2 * 3.0)
        else:
            pen.setWidthF(3.0)
        ps.setCapStyle(pen.capStyle())
        ps.setWidth(pen.widthF() + 10)
        ps.setJoinStyle(pen.joinStyle())
        ps.setMiterLimit(pen.miterLimit())
        p = ps.createStroke(self._path)
        for s in self._src_pts:
            p.addEllipse(s, 3.0, 3.0)
        for d in self._dst_pts:
            p.addEllipse(d, 5.0, 5.0)
        return p

    def boundingRect(self) -> QRectF:
        return self.shape().boundingRect()

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget: Union[QWidget, None] = ...) -> None:
        painter.save()
        # hovered
        if self._hovered or self.isSelected():
            p = QPen()
            p.setWidthF(2 * 3.0)
            p.setColor(QColor(255, 165, 0) if self.isSelected()
                       else QColor(224, 255, 255))
            painter.setPen(p)
            painter.setBrush(Qt.BrushStyle.NoBrush)
            painter.drawPath(self._path)
        # normal
        p = QPen()
        p.setWidthF(3.0)
        if len(self._src_pts) > 1:
            p.setColor(QColor(248, 4, 2))
        else:
            p.setColor(QColor(100, 100, 100) if self.isSelected()
                       else QColor(0, 139, 139))
        painter.setPen(p)
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawPath(self._path)
        # ellipse
        for s in self._src_pts:
            painter.setPen(QColor(128, 128, 128))
            painter.setBrush(QColor(128, 128, 128))
            painter.drawEllipse(s, 3.0, 3.0)
        for d in self._dst_pts:
            painter.setPen(QColor(128, 128, 128))
            painter.setBrush(QColor(128, 128, 128))
            painter.drawEllipse(d, 5.0, 5.0)
        painter.restore()

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        # print('hoverEnterEvent')
        self._hovered = True
        self.update()
        return super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        # print('hoverLeaveEvent')
        self._hovered = False
        self.update()
        return super().hoverLeaveEvent(event)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'datainspector.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QHBoxLayout, QLabel, QLineEdit, QPushButton,
    QSizePolicy, QSpacerItem, QTabWidget, QToolButton,
    QVBoxLayout, QWidget)

class Ui_DataInspector(object):
    def setupUi(self, DataInspector):
        if not DataInspector.objectName():
            DataInspector.setObjectName(u"DataInspector")
        DataInspector.resize(400, 300)
        self.verticalLayout = QVBoxLayout(DataInspector)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.tabWidget = QTabWidget(DataInspector)
        self.tabWidget.setObjectName(u"tabWidget")
        self.from_file = QWidget()
        self.from_file.setObjectName(u"from_file")
        self.verticalLayout_2 = QVBoxLayout(self.from_file)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.verticalLayout_3 = QVBoxLayout()
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.name_label = QLabel(self.from_file)
        self.name_label.setObjectName(u"name_label")

        self.verticalLayout_3.addWidget(self.name_label)

        self.label = QLabel(self.from_file)
        self.label.setObjectName(u"label")

        self.verticalLayout_3.addWidget(self.label)


        self.horizontalLayout_2.addLayout(self.verticalLayout_3)

        self.verticalLayout_4 = QVBoxLayout()
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.var_name = QLineEdit(self.from_file)
        self.var_name.setObjectName(u"var_name")

        self.verticalLayout_4.addWidget(self.var_name)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.path_edit = QLineEdit(self.from_file)
        self.path_edit.setObjectName(u"path_edit")

        self.horizontalLayout.addWidget(self.path_edit)

        self.file_btn = QToolButton(self.from_file)
        self.file_btn.setObjectName(u"file_btn")

        self.horizontalLayout.addWidget(self.file_btn)


        self.verticalLayout_4.addLayout(self.horizontalLayout)


        self.horizontalLayout_2.addLayout(self.verticalLayout_4)


        self.verticalLayout_2.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.load_btn = QPushButton(self.from_file)
        self.load_btn.setObjectName(u"load_btn")

        self.horizontalLayout_3.addWidget(self.load_btn)

        self.dump_to_btn = QPushButton(self.from_file)
        self.dump_to_btn.setObjectName(u"dump_to_btn")

        self.horizontalLayout_3.addWidget(self.dump_to_btn)


        self.verticalLayout_2.addLayout(self.horizontalLayout_3)

        self.data_info = QLabel(self.from_file)
        self.data_info.setObjectName(u"data_info")

        self.verticalLayout_2.addWidget(self.data_info)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer_2)

        self.tabWidget.addTab(self.from_file, "")

        self.verticalLayout.addWidget(self.tabWidget)

        self.buttonBox = QDialogButtonBox(DataInspector)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(DataInspector)
        self.buttonBox.accepted.connect(DataInspector.accept)
        self.buttonBox.rejected.connect(DataInspector.reject)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(DataInspector)
    # setupUi

    def retranslateUi(self, DataInspector):
        DataInspector.setWindowTitle(QCoreApplication.translate("DataInspector", u"Dialog", None))
        self.name_label.setText(QCoreApplication.translate("DataInspector", u"Name:", None))
        self.label.setText(QCoreApplication.translate("DataInspector", u"Load Path:", None))
        self.file_btn.setText(QCoreApplication.translate("DataInspector", u"...", None))
        self.load_btn.setText(QCoreApplication.translate("DataInspector", u"Load", None))
        self.dump_to_btn.setText(QCoreApplication.translate("DataInspector", u"Dump to ...", None))
        self.data_info.setText(QCoreApplication.translate("DataInspector", u"TextLabel", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.from_file), QCoreApplication.translate("DataInspector", u"From File", None))
    # retranslateUi




from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import QGraphicsScene, QMenu, QGraphicsSceneContextMenuEvent, QDialog
from PySide6.QtCore import Signal, Slot, Qt, QRectF, QPointF
from ...ir import Graph, Node, Variable
from .node import GraphNode
from .normal_node import NormalGraphNode
from .io_node import IOGraphNode
from .edge import GraphEdge
from typing import List, Union
from grandalf.graphs import graph_core as GC
from grandalf.graphs import Graph as GG
from grandalf.graphs import Vertex as NN
from grandalf.graphs import Edge as EE
from grandalf.layouts import SugiyamaLayout
from ..ui import IOSummary, NodeSummary, DataInspector
import time
from typing import Union


class GraphScene(QGraphicsScene):
    def __init__(self, ir: Graph, parent=None):
        super().__init__(parent)
        self._ir: Graph = ir
        self._normal_node = []
        self._io_node = []
        self._edge = []
        if self._ir is not None:
            def del_item(o: Union[Node, Variable]):
                n: GraphNode = o.read_ext('bind_gnode')
                n.setPos(0, 0)
                self.removeItem(n)
            self._ir.var_add_callback = self.bind_edge
            self._ir.node_add_callback = self.bind_node
            self._ir.node_del_callback = del_item
            self._ir.var_mark_input_callback = self.bind_node
            self._ir.var_mark_output_callback = self.bind_node
            self._ir.var_unmark_input_callback = del_item
            self._ir.var_unmark_output_callback = del_item
            # init var before node
            # because signal connect in node
            for v in self._ir.variables:
                self.bind_edge(v)
            for v in self._ir.input:
                self.bind_node(v)
            for v in self._ir.output:
                self.bind_node(v)
            for n in self._ir.nodes:
                self.bind_node(n)

    def bind_node(self, ir: Union[Node, Variable]):
        if isinstance(ir, Node):
            if ir.op_type in ['Constant']:
                # we will skip display constant
                return None
            n = NormalGraphNode(ir, self)
            self._normal_node.append(n)
        else:
            assert isinstance(ir, Variable)
            n = IOGraphNode(ir, self)
            self._io_node.append(n)
        assert ir.read_ext('bind_gnode') is None
        ir.set_ext('bind_gnode', n)
        self.addItem(n)
        n.connectToEdge()
        return n

    def bind_edge(self, ire: Variable):
        e = GraphEdge(ire, self)
        self._edge.append(e)
        assert ire.read_ext('bind_gedge') is None
        ire.set_ext('bind_gedge', e)
        self.addItem(e)
        return e

    def layout(self):
        ts = time.time()
        print('cvt to layout ir')
        N = {n.ir.id: NN(n) for n in self._normal_node}
        N.update({n.ir.id: NN(n) for n in self._io_node})
        if len(N) == 0:
            print('skip layout because empty graph')
            return (None, None)
        E = []
        for n in N.values():
            n = n.data
            if isinstance(n, NormalGraphNode):
                src = n.ir.prevNodes
                dst = n.ir.nextNodes
                for v in n.ir.input:
                    if v.isInput:
                        src.append(v)
                for v in n.ir.output:
                    if v.isOutput:
                        dst.append(v)
            elif isinstance(n, IOGraphNode):
                src = n.ir.src
                dst = n.ir.dst
            else:
                raise RuntimeError(f'Unexcept type: {type(n)}')
            for s in src:
                if s.id in N:
                    E.append(
                        EE(N[s.id], N[n.ir.id], data=f'{s.name} -> {n.ir.name}'))
            for d in dst:
                if d.id in N:
                    E.append(
                        EE(N[n.ir.id], N[d.id], data=f'{n.ir.name} -> {d.name}'))
        print('cvt done:', time.time() - ts, 's')
        ts = time.time()
        print('gen gc')
        g = GG(list(N.values()), E)
        print('gen done:', time.time() - ts, 's')
        print('graph_core num:', len(g.C))
        ts = time.time()
        print('layout start')
        for n in N.values():
            rect = n.data.boundingRect()
            assert rect is not None

            class HWView(object):
                w, h = rect.width(), rect.height()
            n.view = HWView()
            # print(n.data.ir.name, n.view.w, n.view.h)
        sug = SugiyamaLayout(g.C[0])
        sug.init_all()
        sug.draw()
        box = QRectF()
        first_node = None
        for n in g.C[0].sV:
            x, y = n.view.xy
            w = n.view.w
            h = n.view.h
            x = x - w / 2
            y = y - h / 2
            this_box = QRectF(x, y, w, h)
            box |= this_box
            # print(n.data.ir.name, x, y)
            n.data.setPos(x, y)
            if first_node is None:
                first_node = n.data
            # print(n.data.ir.name, n.data.pos())
        print('layout done:', time.time() - ts, 's')
        return (box, first_node)

    def contextMenuEvent(self, event: QGraphicsSceneContextMenuEvent) -> None:
        super().contextMenuEvent(event)

        def add_node():
            n = self.addNodeDialog()
            if n is not None:
                n.setPos(event.scenePos())

        def add_io(input):
            def f():
                n = self.addIODialog(input)
                if n is not None:
                    n.setPos(event.scenePos())
            return f
        m = QMenu()
        new_node_act = m.addAction("New Node")
        new_node_act.triggered.connect(add_node)
        new_input_act = m.addAction("New Input")
        new_input_act.triggered.connect(add_io(True))
        new_output_act = m.addAction("New Output")
        new_output_act.triggered.connect(add_io(False))
        add_constant_act = m.addAction("Add Constant")
        add_constant_act.triggered.connect(self.addConstantDialog)
        m.exec(event.screenPos())
        event.accept()

    def addNode(self, name, op_type, inputs, outputs, attrs):
        n = self._ir.addNode(name, op_type)
        n.input = inputs
        n.output = outputs
        n.clearAttr()
        for k, v in attrs.items():
            n.setAttr(k, v)
        assert n.read_ext('bind_gnode') is not None
        return n.read_ext('bind_gnode')

    def addIO(self, name, shape, type, input):
        v = self._ir.getVariable(name)
        v.shape = shape
        v.type = type
        if input:
            self._ir.markInput(v)
        else:
            self._ir.markOutput(v)
        assert v.read_ext('bind_gedge') is not None
        assert v.read_ext('bind_gnode') is not None
        return v.read_ext('bind_gnode')

    def delNode(self, n: NormalGraphNode):
        assert isinstance(n, NormalGraphNode)
        n.ir.removeFromGraph()
        # below was skip because we using update edge by graph ir callback
        # n.setPos(0, 0)
        # self.removeItem(n)

    def delIO(self, n: IOGraphNode):
        assert isinstance(n, IOGraphNode)
        if n.ir.isInput:
            n.ir.unMarkInput()
        if n.ir.isOutput:
            n.ir.unMarkOutput()
        # below was skip because we using update edge by graph ir callback
        # n.setPos(0, 0)
        # self.removeItem(n)

    def addNodeDialog(self):
        dialog = NodeSummary()
        dialog.setWindowTitle('Add Node')
        ret = dialog.exec()
        if ret == QDialog.DialogCode.Accepted:
            ret = dialog.getRet()
            return self.addNode(ret['name'], ret['op_type'], ret['inputs'], ret['output'], ret['attrs'])
        return None

    def addIODialog(self, input=True):
        dialog_name = 'Input' if input else 'Output'
        dialog = IOSummary()
        dialog.setWindowTitle('Add ' + dialog_name)
        ret = dialog.exec()
        if ret == QDialog.DialogCode.Accepted:
            ret = dialog.getRet()
            return self.addIO(ret['name'], ret['shape'], ret['type'], input)
        return None

    def addConstantDialog(self):
        dialog = DataInspector()
        dialog.setWindowTitle('Add Constant')
        ret = dialog.exec()
        if ret == QDialog.DialogCode.Accepted:
            data, name = dialog.getRet()
            self._ir.getVariable(name).data = data

    def keyPressEvent(self, event: QKeyEvent) -> None:
        super().keyPressEvent(event)
        if event.isAccepted():
            return
        if event.key() in [Qt.Key.Key_Delete, Qt.Key.Key_Backspace]:
            for item in self.selectedItems():
                if isinstance(item, NormalGraphNode):
                    self.delNode(item)
                elif isinstance(item, IOGraphNode):
                    self.delIO(item)
